/*
	v9t9 to mess image converter.

	Raphael Nabet 2003
*/

#if defined(__MWERKS__) && macintosh
#include <console.h>
#endif

#include <stdio.h>
#include <string.h>

typedef unsigned char UINT8;
typedef unsigned short UINT16;

#define MAX_SECTOR_LEN 256

/*
	Disk geometry
*/
typedef struct ti99_geometry
{
	int secspertrack;
	int tracksperside;
	int sides;
} ti99_geometry;


typedef struct ti99_vib
{
	char	name[10];			/* volume name (10 characters, pad with spaces) */
	UINT8	totsecsMSB;			/* disk length in sectors (big-endian) (usually 360, 720 or 1440) */
	UINT8	totsecsLSB;
	UINT8	secspertrack;		/* sectors per track (usually 9 (FM) or 18 (MFM)) */
	UINT8	id[3];				/* 'DSK' */
	UINT8	protection;			/* 'P' if disk is protected, ' ' otherwise. */
	UINT8	tracksperside;		/* tracks per side (usually 40) */
	UINT8	sides;				/* sides (1 or 2) */
	UINT8	density;			/* 1 (FM) or 2 (MFM) */
	UINT8	res[36];			/* reserved */
	UINT8	abm[200];			/* allocation bitmap: a 1 for each sector in use (sector 0 is LSBit of byte 0, sector 7 is MSBit of byte 0, sector 8 is LSBit of byte 1, etc.) */
} ti99_vib;


static long flenght(FILE *file)
{
	long pos_save, reply;

	pos_save = ftell(file);
	fseek(file, 0, SEEK_END);
	reply = ftell(file);
	fseek(file, pos_save, SEEK_END);

	return reply;
}

static int read_format(FILE *in, ti99_geometry *geometry)
{
	ti99_vib vib;
	int totsecs;

	/*if (mame_fseek(in, 0, SEEK_SET))
		return 1;*/

	if (fread(& vib, sizeof(vib), 1, in) != 1)
		return 1;

	/* If we have read the sector successfully, let us parse it */
	totsecs = (vib.totsecsMSB << 8) | vib.totsecsLSB;
	geometry->secspertrack = vib.secspertrack;
	if (geometry->secspertrack == 0)
		/* Some images might be like this, because the original SSSD TI
		controller always assumes 9. */
		geometry->secspertrack = 9;
	geometry->tracksperside = vib.tracksperside;
	if (geometry->tracksperside == 0)
		/* Some images are like this, because the original SSSD TI
		controller always assumes 40. */
		geometry->tracksperside = 40;
	geometry->sides = vib.sides;
	if (geometry->sides == 0)
		/* Some images are like this, because the original SSSD TI
		controller always assumes that tracks beyond 40 are on side 2. */
		geometry->sides = totsecs / (geometry->secspertrack * geometry->tracksperside);
	/*geometry->density = vib.density;
	if (geometry->density == 0)
		geometry->density = 1;*/
	/* check that the format makes sense */
	if (((geometry->secspertrack * geometry->tracksperside * geometry->sides) != totsecs)
		/*|| (geometry->density > 3)*/ || (totsecs < 2) || memcmp(vib.id, "DSK", 3)
		|| (flenght(in) != totsecs*256))
	{
		/* invalid geometry */
		return 1;
	}

	return 0;
}

int main(int argc, char *argv[])
{
	FILE *in, *out;
	ti99_geometry geometry;
	int cylinder, side, sector;
	unsigned int offset;
	char buf[256];


#if defined(__MWERKS__) && macintosh
	argc = ccommand(& argv);
#endif

	if (argc != 3)
	{
		fprintf(stderr, "%s source dest\n", argv[0]);
		return 1;
	}

	in = fopen(argv[1], "rb");
	if (in == NULL)
	{
		fprintf(stderr, "can't open source\n");
		return 1;
	}

	out = fopen(argv[2], "wb");
	if (out == NULL)
	{
		fprintf(stderr, "can't open dest\n");
		fclose(in);
		return 1;
	}

	if (read_format(in, &geometry))
	{
		fprintf(stderr, "invalid source image\n");
		fclose(out);
		fclose(in);
		return 1;
	}

	if (geometry.sides == 1)
		fprintf(stderr, "source image is single-sided: it is compatible with both MESS and V9T9\n");

	rewind(in);

	for (side = 0; side < geometry.sides; side++)
		for (cylinder = ((side & 1) ? (geometry.tracksperside-1) : 0);
				(side & 1) ? (cylinder >= 0) : (cylinder < geometry.tracksperside);
				cylinder = (side & 1) ? (cylinder - 1) : (cylinder + 1))
		{
			/* compute offset in mess image */
			offset = (cylinder*geometry.sides + side)*geometry.secspertrack*256;
			fseek(out, offset, SEEK_SET);
			for (sector = 0; sector < geometry.secspertrack; sector++)
			{
				/* read sector */
				if (fread(buf, 256, 1, in) != 1)
				{
					fprintf(stderr, "read error in source\n");
					fclose(out);
					fclose(in);
					return 1;
				}
				if (fwrite(buf, 256, 1, out) != 1)
				{
					fprintf(stderr, "write error in dest\n");
					fclose(out);
					fclose(in);
					return 1;
				}
			}
		}

	fclose(out);
	fclose(in);

	fprintf(stderr, "done\n");

	return 0;
}
